# The performance of the methods and the baselines given the number of sources
# on multiple graphs: Planetoid, ER, and SW

import warnings

warnings.filterwarnings("ignore")

from methods import *
import time
import statistics as s

print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
print(diffusion_model)

################################################
# Global parameters
################################################
num_of_sims = 10

graphs = ['synn','SW','Cora','CiteSeer', 'PubMed'] # 'ER', 'BA',

for graph_type in graphs:
    print('============================================')
    if graph_type == 'synn':
        print('synn')
        G, config = synn()
    if graph_type == 'SW':
        print('small world')
        G, config = connSW(3000)
    elif graph_type == 'ER':
        print('ER')
        G, config = ER(3000)
    elif graph_type == 'BA':
        print('BA')
        G, config = BA(3000)
    elif graph_type == 'Cora':
        print('Cora')
        G, config = Cora()
    elif graph_type == 'CiteSeer':
        print('CiteSeer')
        G, config = CiteSeer()
    elif graph_type == 'PubMed':
        print('PubMed')
        G, config = PubMed()

    BOIM = []
    BOIM_wo_filtering = []
    BOIM_wo_fourier = []
    BOIM_wo_both = []
    BOIM_n = []
    deg = []
    eig = []
    dd = []
    pi_method = []
    sigma_method = []
    Soboldegree = []



    for i in range(5):
        
        result = BOIM_full(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources, allowed_shortest_distance, number_of_clusters)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        BOIM.append(spread)

        result = BOIM_no_filtering(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        BOIM_wo_filtering.append(spread)

        result = BOIM_no_fourier(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources, allowed_shortest_distance)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        BOIM_wo_fourier.append(spread)

        result = BOIM_vanilla(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        BOIM_wo_both.append(spread)

        result = BOIM_no_GSS(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources, allowed_shortest_distance)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        BOIM_n.append(spread)

        result = degreeDis(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        dd.append(spread)

        result = degree(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        deg.append(spread)

        result = eigen(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        eig.append(spread)

        result = pi(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        pi_method.append(spread)

        result = sigma(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        sigma_method.append(spread)

        result = SoboldegreeDis(G, config, number_of_sources)
        if diffusion_model == 'ic':
            spread,_ = effectIC(G, config, result, rounds = num_of_sims)
        elif diffusion_model == 'lt':
            spread,_ = effectLT(G, config, result, rounds = num_of_sims)
        Soboldegree.append(spread)

    print('BOIM full')
    print('eval: ', s.mean(BOIM), '+-', s.stdev(BOIM))

    print('BOIM no filtering')
    print('eval: ', s.mean(BOIM_wo_filtering), '+-', s.stdev(BOIM_wo_filtering))

    print('BOIM no fourier')
    print('eval: ', s.mean(BOIM_wo_fourier), '+-', s.stdev(BOIM_wo_fourier))

    print('BOIM vanilla')
    print('eval: ', s.mean(BOIM_wo_both), '+-', s.stdev(BOIM_wo_both))

    print('BOIM no GSS')
    print('eval: ', s.mean(BOIM_n), '+-', s.stdev(BOIM_n))

    print('degree discount')
    print('eval: ', s.mean(dd), '+-', s.stdev(dd))

    print('degree')
    print('eval: ', s.mean(deg), '+-', s.stdev(deg))

    print('eigen')
    print('eval: ', s.mean(eig), '+-', s.stdev(eig))

    print('pi')
    print('eval: ', s.mean(pi_method), '+-', s.stdev(pi_method))

    print('sigma')
    print('eval: ', s.mean(sigma_method), '+-', s.stdev(sigma_method))

    print('Sobol degree discount')
    print('eval: ', s.mean(Soboldegree), '+-', s.stdev(Soboldegree))

