# The scalability of the methods and the baselines
# evaluated on different sizes of SW graphs

import warnings
import gc

warnings.filterwarnings("ignore")

from my_util import *
from methods import *
import time
import statistics as s

print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

################################################
# Global parameters
################################################

for n in [1000,2000,3000,4000,5000]:
    gc.collect()

    print('===============================================================')
    print('n = ', n)

    BOIMT = []
    SlowT = []
    degT = []
    eigT = []
    degreeDisT = []
    PiT = []
    SigmaT = []
    sobolT = []

    for i in range(5):

        # generate a SW graph
        G, config = connSW(n)

        ################################################
        # methods evaluation
        ################################################
        start = time.time()
        result = BOIM_no_filtering(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources)
        end = time.time()
        SlowT.append(end - start)

        start = time.time()
        result = BOIM_no_GSS(G, config, num_iterations, num_of_sims, candidate_size, diffusion_model, number_of_sources, allowed_shortest_distance)
        end = time.time()
        BOIMT.append(end - start)

        start = time.time()
        result = degreeDis(G, config, number_of_sources)
        end = time.time()
        degreeDisT.append(end - start)

        start = time.time()
        result = degree(G, config, number_of_sources)
        end = time.time()
        degT.append(end - start)

        start = time.time()
        result = eigen(G, config, number_of_sources)
        end = time.time()
        eigT.append(end - start)

        start = time.time()
        result = pi(G, config, number_of_sources)
        end = time.time()
        PiT.append(end - start)

        start = time.time()
        result = sigma(G, config, number_of_sources)
        end = time.time()
        SigmaT.append(end - start)

        start = time.time()
        result = SoboldegreeDis(G, config, number_of_sources)
        end = time.time()
        sobolT.append(end - start)        

    print('BOIM: ', s.mean(BOIMT), '+-', s.stdev(BOIMT))
    print('BOIM new: ', s.mean(SlowT), '+-', s.stdev(SlowT))
    print('Degree Discount: ', s.mean(degreeDisT), '+-', s.stdev(degreeDisT))
    print('Degree: ', s.mean(degT), '+-', s.stdev(degT))
    print('Eigen: ', s.mean(eigT), '+-', s.stdev(eigT))
    print('Pi: ', s.mean(PiT), '+-', s.stdev(PiT))
    print('Sigma: ', s.mean(SigmaT), '+-', s.stdev(SigmaT))
    print('Sobol degree discount: ', s.mean(sobolT), '+-', s.stdev(sobolT))

    gc.collect()
